DASHBOARD_API = "/dashboard"
VALIDATE_KEY_API = "/api/key/validate"
USERS_LIST_GET_API = "/api/users"
USERS_COUNT_API = "/api/users/count"
USER_API = "/api/user"
USER_EMAIL_VERIFY_API = "/api/user/email/verify"
USER_METADATA_API = "/api/user/metadata"
USER_SESSION_API = "/api/user/sessions"
USER_PASSWORD_API = "/api/user/password"
USER_EMAIL_VERIFY_TOKEN_API = "/api/user/email/verify/token"
EMAIL_PASSWORD_SIGN_IN = "/api/signin"
EMAIL_PASSSWORD_SIGNOUT = "/api/signout"
